# SPDX-FileCopyrightText: 2024-present Alex Rudy <github@alexrudy.net>
#
# SPDX-License-Identifier: MIT
from ._version import __version__ as version

author = "Alex Rudy"
email = "opensource@alexrudy.net"

__all__ = ["author", "email", "version"]



from typing import ClassVar

import attrs
from domilite import svg as svg_tag
from domilite.dom_tag import dom_tag

from bootlace.endpoint import Endpoint
from bootlace.util import Tag

__all__ = ["Icon"]


@attrs.define
class Icon:
    """A Bootstrap icon

    This class supports the :func:`as_tag` protocol to display itself.
    """

    #: Endpoint name for getting the Bootstrap Icon SVG file
    endpoint: ClassVar[Endpoint] = Endpoint.from_name("bootlace.static", filename="icons/bootstrap-icons.svg")

    #: Name of the icon
    name: str

    svg: Tag = Tag(
        svg_tag.svg,
        attributes={
            "role": "img",
            "fill": "currentColor",
            "width": "16",
            "height": "16",
        },
        classes={"bi", "me-1", "pe-none", "align-self-center", "bi-inline"},
    )

    use: Tag = Tag(svg_tag.use)

    @property
    def url(self) -> str:
        """The URL for the SVG source for the icon"""
        return self.endpoint(_anchor=self.name)

    def __tag__(self) -> dom_tag:
        return self.svg(
            self.use(xlink_href=self.url),
        )



from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Any

import attrs
from domilite import tags
from markupsafe import Markup
from marshmallow import Schema
from marshmallow import ValidationError
from marshmallow import fields
from marshmallow import post_load
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from bootlace import links
from bootlace.endpoint import Endpoint
from bootlace.util import MaybeTaggable
from bootlace.util import Tag
from bootlace.util import render

if TYPE_CHECKING:
    from bootlace.nav.core import NavElement


class NavSchema(OneOfSchema):
    """Registry for nav element schemas"""

    type_field = "__type__"
    type_schemas: dict[str, type[Schema]] = {}  # pyright: ignore

    def __init__(self, **kwargs: Any) -> None:
        self.register_all()
        super().__init__(**kwargs)

    @classmethod
    def register_all(cls) -> None:
        """Register a schema for a nav element"""
        from bootlace.nav.core import NavElement

        for element in NavElement._NAV_ELEMENT_REGISTRY.values():
            cls.register(element)

    @classmethod
    def register(cls, element: "type[NavElement]") -> None:
        """Register a schema for a nav element"""
        if element.__name__ not in cls.type_schemas:
            schema = build_schema(element)
            cls.type_schemas[element.__name__] = schema


class BaseSchema(Schema):
    """Base schema with reloading"""

    @post_load
    def make_instance(self, data: dict[str, Any], **kwargs: Any) -> Any:
        return self.Meta.element(**data)  # type: ignore


class DomTagField(fields.Field):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs: Any) -> Any:
        if value is None:  # pragma: no cover
            return None

        if not (isinstance(value, tags.html_tag) or issubclass(value, tags.html_tag)):  # pragma: no cover
            raise ValidationError(f"Unknown tag type: {type(value)!r}: {value!r}")

        return value.name

    def _deserialize(
        self,
        value: Any,
        attr: str | None,
        data: Mapping[str, Any] | None,
        **kwargs: Any,
    ) -> Any:
        if value is None:  # pragma: no cover
            return None
        return tags.html_tag.find_tag_type(value)


class Set(fields.List):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs: Any) -> Any:
        if value is None:  # pragma: no cover
            return None
        return list(value)

    def _deserialize(
        self,
        value: Any,
        attr: str | None,
        data: Mapping[str, Any] | None,
        **kwargs: Any,
    ) -> Any:
        if value is None:  # pragma: no cover
            return None
        return set(value)


class TagSchema(Schema):
    tag = DomTagField()
    classes = Set(fields.String())
    attributes = fields.Dict()

    class Meta:
        element = Tag

    @post_load
    def make_instance(self, data: dict[str, Any], **kwargs: Any) -> Any:
        return Tag(**data)


class TaggableField(fields.Field):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs: Any) -> Any:
        if value is None:  # pragma: no cover
            return None
        return str(render(value))

    def _deserialize(
        self,
        value: Any,
        attr: str | None,
        data: Mapping[str, Any] | None,
        **kwargs: Any,
    ) -> Any:
        if value is None:  # pragma: no cover
            return None
        return Markup(value)


ATTRS_FIELD_TYPE_MAP = {
    str: fields.String,
    str | None: lambda: fields.String(allow_none=True),
    int: fields.Integer,
    float: fields.Float,
    bool: fields.Boolean,
    Tag: lambda: fields.Nested(TagSchema),
    links.LinkBase: lambda: fields.Nested(LinkBaseSchema),
    MaybeTaggable: TaggableField,
}


def build_schema(element: "type[NavElement]") -> type[Schema]:
    form_fields: dict[str, fields.Field] = {}
    for field in attrs.fields(element):  # type: ignore
        ff = field.metadata.get("form", None)
        if ff is None:
            if field.type not in ATTRS_FIELD_TYPE_MAP:  # pragma: no cover
                raise ValueError(f"Unknown field type {field.type!r}")

            form_fields[field.name] = ATTRS_FIELD_TYPE_MAP[field.type]()  # type: ignore
        else:
            form_fields[field.name] = ff

    meta = type("Meta", (), {"element": element})

    attributes: dict[str, Any] = {"Meta": meta, **form_fields}

    return type(f"{element.__name__}Schema", (BaseSchema,), attributes)


class EndpointSchema(BaseSchema):
    name = fields.String(required=True)
    url_kwargs = fields.Dict()
    ignore_query = fields.Boolean(required=True)

    class Meta:
        element = Endpoint


class LinkSchema(BaseSchema):
    text = TaggableField()
    a = fields.Nested(TagSchema)
    url = fields.String(required=True)
    active = fields.Boolean(required=True)
    enabled = fields.Boolean(required=True)

    class Meta:
        element = links.Link


class ViewLinkSchema(BaseSchema):
    text = TaggableField()
    a = fields.Nested(TagSchema)
    endpoint = fields.Nested(EndpointSchema)
    enabled = fields.Boolean()

    class Meta:
        element = links.View


class LinkBaseSchema(OneOfSchema):
    type_field = "__type__"
    type_schemas: dict[str, type[Schema]] = {"Link": LinkSchema, "View": ViewLinkSchema}  # pyright: ignore



from .bootstrap import Switch
from .core import CheckboxInput
from .core import FileInput
from .core import HiddenInput
from .core import Input
from .core import PasswordInput
from .core import RadioInput
from .core import SubmitInput
from .core import TextArea
from .core import TextInput
from .select import Select
from .typed import ColorInput
from .typed import DateInput
from .typed import DateTimeInput
from .typed import DateTimeLocalInput
from .typed import EmailInput
from .typed import MonthInput
from .typed import NumberInput
from .typed import RangeInput
from .typed import SearchInput
from .typed import TelInput
from .typed import TimeInput
from .typed import URLInput
from .typed import WeekInput

__all__ = [
    "Input",
    "TextInput",
    "PasswordInput",
    "HiddenInput",
    "CheckboxInput",
    "RadioInput",
    "FileInput",
    "SubmitInput",
    "TextArea",
    "Select",
    "SearchInput",
    "TelInput",
    "URLInput",
    "EmailInput",
    "DateTimeInput",
    "DateInput",
    "MonthInput",
    "WeekInput",
    "TimeInput",
    "DateTimeLocalInput",
    "NumberInput",
    "RangeInput",
    "ColorInput",
    "Switch",
]



from typing import Any

import attrs
from flask import Blueprint
from flask import Flask
from flask import current_app

from .endpoint import Endpoint
from .resources import Resources
from .util import as_tag
from .util import render


@attrs.define(init=False, eq=False)
class Bootlace:
    """Flask extension for bootlace"""

    def __init__(self, app: Flask | None = None) -> None:
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Initialize the extension with the Flask app"""

        app.extensions["bootlace"] = self
        app.jinja_env.globals["bootlace"] = self

        name = app.config.setdefault("BOOTLACE_BLUEPRINT_NAME", "bootlace")

        if app.config.setdefault("BOOTLACE_CONTEXT_PROCESSORS", True):
            app.context_processor(context)

        blueprint = Blueprint(
            name,
            __name__,
            template_folder="templates",
            static_folder="static",
            static_url_path="/static/bootstrap",
        )

        app.register_blueprint(blueprint)

    @property
    def static_view(self) -> str:
        bp = current_app.config["BOOTLACE_BLUEPRINT_NAME"]
        return f"{bp}.static"

    def bootstrap(self) -> "Resources":
        return Resources(
            endpoint=Endpoint.from_name(self.static_view),
            resources=["bootstrap.min.js", "bootstrap.min.css"],
        )


def context() -> dict[str, Any]:
    return {
        "render": render,
        "as_tag": as_tag,
    }



# SPDX-FileCopyrightText: 2024-present Alex Rudy <github@alexrudy.net>
#
# SPDX-License-Identifier: MIT
from . import __about__
from ._version import __version__
from .extension import Bootlace
from .util import as_tag
from .util import render

__all__ = ["__version__", "__about__", "as_tag", "render", "Bootlace"]



from collections.abc import Iterable
from collections.abc import Iterator
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import NamedTuple
from typing import TypeVar

import attrs
from domilite import tags
from markupsafe import escape
from wtforms.fields import Field
from wtforms.fields import SelectField

from bootlace.util import Tag

from .core import Widget

__all__ = ["Select", "OptionChoice"]

V = TypeVar("V")


class OptionChoice(NamedTuple, Generic[V]):
    value: V
    label: str
    selected: bool
    render_kw: dict[str, Any]

    @classmethod
    def map(cls, iterator: Iterable[tuple[V, str, bool, dict[str, Any]]]) -> "Iterator[OptionChoice]":
        for item in iterator:
            yield cls(*item)


@attrs.define
class Select(Widget):
    multiple: bool = False
    validation_attrs = {"required", "disabled"}
    tag = Tag(tags.select, classes={"form-control"})
    optgroup: ClassVar[Tag] = Tag(tags.optgroup)
    option: ClassVar[Tag] = Tag(tags.option)

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if not isinstance(field, SelectField):
            raise TypeError(f"Field for select widget must be a SelectField, got {type(field)!r}")
        if self.multiple:
            kwargs["multiple"] = True
        widget_tag = super().__form_tag__(field, **kwargs)

        if field.has_groups():
            for group, choices in field.iter_groups():
                group_tag = self.optgroup(label=group)
                for choice in OptionChoice.map(choices):
                    group_tag.add(self.render_option(choice))
                widget_tag.add(group_tag)
        else:
            for choice in OptionChoice.map(field.iter_choices()):
                widget_tag.add(self.render_option(choice))
        return widget_tag

    def render_option(self, choice: OptionChoice) -> tags.html_tag:
        if choice.value is True:
            value = str(choice.value)
        else:
            value = choice.value
        options = dict(choice.render_kw, value=value)
        if choice.selected:
            options["selected"] = True
        return self.option(escape(choice.label), **options)



import abc

import attrs
from domilite import tags

from .endpoint import Endpoint
from .endpoint import convert_endpoint
from .util import MaybeTaggable
from .util import Tag
from .util import as_tag

__all__ = ["Link", "View"]


@attrs.define(kw_only=True, frozen=True)
class LinkBase(abc.ABC):
    text: MaybeTaggable

    a: Tag = Tag(tags.a)

    @property
    @abc.abstractmethod
    def active(self) -> bool:
        raise NotImplementedError("LinkBase.active must be implemented in a subclass")

    @property
    @abc.abstractmethod
    def enabled(self) -> bool:
        raise NotImplementedError("LinkBase.enabled must be implemented in a subclass")

    @property
    @abc.abstractmethod
    def url(self) -> str:
        raise NotImplementedError("LinkBase.url must be implemented in a subclass")

    def __tag__(self) -> tags.html_tag:
        return self.a(as_tag(self.text), href=self.url)


@attrs.define(kw_only=True, frozen=True)
class Link(LinkBase):
    """A raw link to a URL."""

    #: The URL to link to
    url: str

    #: Whether the link is active
    active: bool = False

    #: Whether the link is enabled
    enabled: bool = True


@attrs.define(kw_only=True, frozen=True)
class View(LinkBase):
    """Link to a Flask view."""

    #: The endpoint to link to, for use with Flask's :func:`~flask.url_for`
    endpoint: Endpoint = attrs.field(converter=convert_endpoint)

    #: Whether the link is enabled.
    enabled: bool = True

    @property
    def url(self) -> str:
        """The URL to link to, constructed using :func:`~flask.url_for`."""
        return self.endpoint.url

    @property
    def active(self) -> bool:
        """Whether the link is active, based on the current request endpoint."""
        return self.endpoint.active

    @property
    def blueprint(self) -> str | None:
        """The blueprint for the endpoint."""
        return self.endpoint.blueprint



import enum
import warnings
from typing import Any

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from marshmallow import fields

from bootlace import links
from bootlace.endpoint import Endpoint
from bootlace.nav.schema import NavSchema
from bootlace.util import BootlaceWarning
from bootlace.util import MaybeTaggable
from bootlace.util import Tag
from bootlace.util import as_tag
from bootlace.util import ids as element_id


class NavStyle(enum.Enum):
    """Styles for the nav element"""

    PLAIN = ""
    TABS = "nav-tabs"
    PILLS = "nav-pills"
    UNDERLINE = "nav-underline"


class NavAlignment(enum.Enum):
    """Alignment for the nav element"""

    DEFAULT = ""
    FILL = "nav-fill"
    JUSTIFIED = "nav-justified"


class NavElement:
    """Base class for nav components"""

    Schema: type[NavSchema] = NavSchema
    _NAV_ELEMENT_REGISTRY: dict[str, type["NavElement"]] = {}

    def __init_subclass__(cls) -> None:
        cls._NAV_ELEMENT_REGISTRY[cls.__name__] = cls

    @property
    def active(self) -> bool:
        """Whether the element is active"""
        return False

    @property
    def enabled(self) -> bool:
        """Whether the element is enabled"""
        return True

    def __tag__(self) -> dom_tag:
        warnings.warn(
            BootlaceWarning(f"Unhandled element {self.__class__.__name__}"),
            stacklevel=2,
        )
        return tags.comment(f"unhandled element {self.__class__.__name__}")

    def element_state(self, tag: dom_tag) -> dom_tag:
        """Apply :attr:`active` and :attr:`enabled` states to the tag."""
        if not isinstance(tag, tags.html_tag):
            return tag

        if self.active:
            tag.classes.add("active")
            tag.aria["current"] = "page"

        if not self.enabled:
            tag.classes.add("disabled")
            tag.aria["disabled"] = "true"
        return tag


@attrs.define
class Link(NavElement):
    """A link in the nav bar, either for a view or a literal URL"""

    #: The link to display, either a URL or a view
    link: links.LinkBase

    #: The ID of the element
    id: str = attrs.field(factory=element_id.factory("nav-link"))

    a: Tag = Tag(tags.a, classes={"nav-link"})

    @classmethod
    def with_url(cls, url: str, text: MaybeTaggable, **kwargs: Any) -> "Link":
        """Create a link with a URL."""
        return cls(link=links.Link(url=url, text=text, **kwargs))

    @classmethod
    def with_view(cls, endpoint: str, text: MaybeTaggable, **kwargs: Any) -> "Link":
        """Create a link with a view."""
        return cls(link=links.View(endpoint=Endpoint.from_name(endpoint, **kwargs), text=text))

    @property
    def active(self) -> bool:
        """Whether the link is active."""
        return self.link.active

    @property
    def enabled(self) -> bool:
        """Whether the link is enabled."""
        return self.link.enabled

    @property
    def url(self) -> str:
        """The URL for the link."""
        return self.link.url

    def __tag__(self) -> dom_tag:
        a = as_tag(self.link)
        if isinstance(a, tags.html_tag):
            a["id"] = self.id
            self.a.update(a)

        return self.element_state(a)


@attrs.define
class Separator(NavElement):
    """A separator in dropdown menus"""

    hr: Tag = Tag(tags.hr, classes={"dropdown-divider"})

    def __tag__(self) -> tags.html_tag:
        return self.hr()


@attrs.define
class Text(NavElement):
    """A text element in the nav bar"""

    text: MaybeTaggable

    span: Tag = Tag(tags.span, classes={"nav-link"})

    @property
    def enabled(self) -> bool:
        return False

    def __tag__(self) -> dom_tag:
        tag = self.span(self.text, cls="nav-link")
        return self.element_state(tag)


@attrs.define
class SubGroup(NavElement):
    """Any grouping of items in the nav bar"""

    items: list[NavElement] = attrs.field(factory=list, metadata={"form": fields.List(fields.Nested(NavSchema))})

    @property
    def active(self) -> bool:
        return any(item.active for item in self.items)



from typing import Any

import attrs
from domilite.tags import html_tag
from domilite.tags import input
from wtforms.fields.core import Field

from bootlace.util import Tag

from .core import InputBase

__all__ = [
    "SearchInput",
    "TelInput",
    "URLInput",
    "EmailInput",
    "DateTimeInput",
    "DateInput",
    "MonthInput",
    "WeekInput",
    "TimeInput",
    "DateTimeLocalInput",
    "NumberInput",
    "RangeInput",
    "ColorInput",
]


@attrs.define
class SearchInput(InputBase):
    """
    Renders an input with type "search".
    """

    input_type = "search"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }


@attrs.define
class TelInput(InputBase):
    """
    Renders an input with type "tel".
    """

    input_type = "tel"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }


@attrs.define
class URLInput(InputBase):
    """
    Renders an input with type "url".
    """

    input_type = "url"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }


@attrs.define
class EmailInput(InputBase):
    """
    Renders an input with type "email".
    """

    input_type = "email"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }


@attrs.define
class DateTimeInput(InputBase):
    """
    Renders an input with type "datetime".
    """

    input_type = "datetime"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class DateInput(InputBase):
    """
    Renders an input with type "date".
    """

    input_type = "date"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class MonthInput(InputBase):
    """
    Renders an input with type "month".
    """

    input_type = "month"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class WeekInput(InputBase):
    """
    Renders an input with type "week".
    """

    input_type = "week"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class TimeInput(InputBase):
    """
    Renders an input with type "time".
    """

    input_type = "time"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class DateTimeLocalInput(InputBase):
    """
    Renders an input with type "datetime-local".
    """

    input_type = "datetime-local"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}


@attrs.define
class NumberInput(InputBase):
    """
    Renders an input with type "number".
    """

    step: int | None = None
    min: int | None = None
    max: int | None = None

    input_type = "number"
    validation_attrs = {"required", "disabled", "readonly", "max", "min", "step"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> html_tag:
        if self.step is not None:
            kwargs.setdefault("step", self.step)
        if self.min is not None:
            kwargs.setdefault("min", self.min)
        if self.max is not None:
            kwargs.setdefault("max", self.max)
        return super().__form_tag__(field, **kwargs)


@attrs.define
class RangeInput(InputBase):
    """
    Renders an input with type "range".
    """

    step: int | None = None
    min: int | None = None
    max: int | None = None

    input_type = "range"
    validation_attrs = {"disabled", "max", "min", "step"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> html_tag:
        if self.step is not None:
            kwargs.setdefault("step", self.step)
        if self.min is not None:
            kwargs.setdefault("min", self.min)
        if self.max is not None:
            kwargs.setdefault("max", self.max)
        return super().__form_tag__(field, **kwargs)


@attrs.define
class ColorInput(InputBase):
    """
    Renders an input with type "color".
    """

    tag = Tag(input, classes={"form-control-color"})

    input_type = "color"
    validation_attrs = {"disabled"}



import enum


class ColorClass(enum.Enum):
    """Bootstrap color classes"""

    #: Default foreground (color) and background, including components.
    BODY = None

    #: Main theme color, used for hyperlinks, focus styles, and component and form active states.
    PRIMARY = "primary"

    #: Secondary theme color, used for secondary buttons and form elements.
    SECONDARY = "secondary"

    #: Tertiary theme color, used for tertiary buttons and form elements.
    TERTIARY = "tertiary"

    #: Success state color, used for success messages and indicators.
    SUCCESS = "success"

    #: Danger state color, used for error messages and indicators.
    DANGER = "danger"

    #: Warning state color, used for warning messages and indicators.
    WARNING = "warning"

    #: Info state color, used for informational messages and indicators.
    INFO = "info"

    #: Additional theme option for less contrasting colors.
    LIGHT = "light"

    #: Additional theme option for higher contrasting colors.
    DARK = "dark"

    def add_to_class(self, cls: str) -> str:
        """Add the color class name to the given class."""
        if self.value:
            return f"{cls}-{self.value}"
        return cls



from typing import Any

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.util import text

from bootlace.endpoint import Endpoint
from bootlace.endpoint import convert_endpoint
from bootlace.icon import Icon
from bootlace.table.base import ColumnBase
from bootlace.util import MaybeTaggable
from bootlace.util import Tag
from bootlace.util import as_tag

__all__ = ["Column", "ActionColumn", "CheckColumn", "Datetime"]


@attrs.define
class Column(ColumnBase):
    """A column in a table, which shows the value of an attribute.

    If no special formatting is applied to the attribute, it is rendered as text.
    """

    format: str | None = attrs.field(default=None)

    def cell(self, value: Any) -> dom_tag:
        """Return the cell for the column as an HTML tag."""
        return self.contents(value)

    def contents(self, value: Any, format: str | None = None) -> dom_tag:
        """Return the contents of the cell for the column, using an HTML comment if the attribute value is None."""
        return super().contents(value, format or self.format)


@attrs.define
class CheckColumn(ColumnBase):
    """A column which shows a checkmark or X based on the value of the attribute."""

    #: The icon for a true value
    yes: Icon = attrs.field(default=Icon("check"))

    #: The icon for a false value
    no: Icon = attrs.field(default=Icon("x"))

    def cell(self, value: Any) -> dom_tag:
        """Return the cell for the column as an HTML tag."""
        if getattr(value, self.attribute):
            return as_tag(self.yes)
        return as_tag(self.no)


@attrs.define
class Datetime(ColumnBase):
    """A column which shows a datetime attribute as an ISO formatted string.

    This column can also be used for date or time objects.

    A format string can be provided to format the datetime object."""

    format: str | None = attrs.field(default=None)

    def cell(self, value: Any) -> dom_tag:
        """Return the cell for the column as an HTML tag."""
        if self.format:
            return text(getattr(value, self.attribute).strftime(self.format))

        return text(getattr(value, self.attribute).isoformat())


@attrs.define
class ActionColumn(ColumnBase):
    """
    A column which links to a view for the value.

    This is commonly shown as e.g. the name of the item, which links to the edit view.
    """

    endpoint: Endpoint = attrs.field(default=Endpoint.from_name(".edit"), converter=convert_endpoint)

    label: MaybeTaggable = attrs.field(default=None)

    a: Tag = Tag(tags.a)

    def cell(self, value: Any) -> tags.html_tag:
        id = getattr(value, "id", None)

        if self.label is None:
            contents = self.contents(value)
        else:
            contents = as_tag(self.label)

        return self.a(contents, href=self.endpoint(id=id))



import collections
import functools
import itertools
import warnings
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from typing import Any
from typing import Protocol
from typing import TypeAlias
from typing import TypeVar

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.template import TagTemplate as Tag
from domilite.util import container
from domilite.util import text
from flask import request
from markupsafe import Markup

T = TypeVar("T")

__all__ = [
    "BootlaceWarning",
    "HtmlIDScope",
    "IntoTag",
    "MaybeTaggable",
    "Taggable",
    "Tag",
    "as_tag",
    "ids",
    "is_active_endpoint",
    "maybe",
    "render",
]


class BootlaceWarning(UserWarning):
    """A warning specific to Bootlace"""


class Taggable(Protocol):
    """Protocol for objects that can be converted to a tag."""

    def __tag__(self) -> dom_tag:
        """Convert the object to a dominate tag.

        This method gives objects control over how they are processed by :func:`as_tag`. It should return a
        :mod:`dominate` tag. If a taggable object contains other taggable objects, it should use :func:`as_tag` to
        convert them, and then apply any additional processing as necessary to the returned :class:`~dominate.html_tag`.

        :meta public:
        :returns: A :mod:`dominate` tag.
        """
        ...


#: A type that can be converted to a tag
IntoTag: TypeAlias = Taggable | dom_tag | str

#: A type that can be converted to a tag via :func:`as_tag`
MaybeTaggable: TypeAlias = IntoTag | str | Iterable[IntoTag]


def as_tag(item: MaybeTaggable) -> dom_tag:
    """Convert an item to a dominate tag.

    :mod:`bootlace` uses :mod:`dominate` to render HTML. To do this, objects implement the :class:`Taggable` protocol,
    providing a ``__tag__`` dunder method. This method will also accept regular :mod:`dominate` tags, strings, and
    iterables of :class:`Taggable` objects. It will try to always return a :mod:`dominate` tag.

    To render taggable objects in a template, use :func:`render`, a convenience function that will convert the object
    to a :mod:`dominate` tag and then render it to a :class:`Markup` object for use in a template.

    Handling notes
    --------------

    When a string is passed in, it will be wrapped with :class:`dominate.util.text` to render a literal string as a
    tag. When an iterable of taggable items is passed, it is returned as a :class:`dominate.util.container`, which will
    render the tags in sequence.

    Unknown types are displayed using their string representation (by calling :class:`str` on them), along with a
    comment in the rendered HTML and a :class:`Bootlace` warning emitted.

    Arguments
    ---------

    :param item: The item to convert to :mod:`dominate` tags.
    :returns: A :mod:`dominate` tag.

    """

    if isinstance(item, tags.html_tag):
        # item.children = [as_tag(child) for child in item.children]
        return item
    if hasattr(item, "__tag__"):
        return item.__tag__()
    if isinstance(item, str):
        return text(item)
    if isinstance(item, Iterable):
        return container(*[as_tag(i) for i in item])

    warnings.warn(
        BootlaceWarning(f"Rendered type {item.__class__.__name__} not explicitly supported"),
        stacklevel=2,
    )
    return container(
        text(str(item)),
        tags.comment(f"Rendered type {item.__class__.__name__} not supported"),
    )


def render(item: MaybeTaggable) -> Markup:
    """Render an item to a Markup object.

    This function is a convenience wrapper around :func:`as_tag` and :meth:`dominate.tags.html_tag.render`. It will try
    to convert most objects to a :mod:`dominate` tag and then render it to a :class:`Markup` object which can be
    inserted into :mod:`jinja` templates.

    Arguments
    ---------
    :param item: The item to render. See :func:`as_tag` for more information.
    :returns: A :class:`Markup` object, suitable for inserting into a :mod:`jinja` template.

    """
    return Markup(as_tag(item).render())


@attrs.define
class HtmlIDScope:
    """A helper for generating unique HTML IDs."""

    #: A mapping of scopes to counters
    scopes: collections.defaultdict[str, itertools.count] = attrs.field(
        factory=lambda: collections.defaultdict(itertools.count)
    )

    def __call__(self, scope: str) -> str:
        """Generate a unique ID for a given scope.

        Parameters
        ----------
        scope : str
            Scopes are used to group IDs together, e.g. items in a list, or a form and its fields.
        """
        counter = next(self.scopes[scope])
        if counter == 0:
            return scope
        return f"{scope}-{counter}"

    def factory(self, scope: str) -> functools.partial:
        """Create a factory function for generating IDs in a specific scope."""
        return functools.partial(self, scope)

    def reset(self) -> None:
        """Reset all ID scopes."""
        self.scopes.clear()


ids = HtmlIDScope()


def maybe(cls: type[T]) -> Callable[[str | T], T]:
    """Convert a string to a class instance if necessary."""

    def converter(value: str | T) -> T:
        if isinstance(value, str):
            return cls(value)  # type: ignore
        return value

    return converter


def is_active_endpoint(endpoint: str, url_kwargs: Mapping[str, Any], ignore_query: bool = True) -> bool:
    """Check if the current request is for the given endpoint and URL kwargs"""
    if request.endpoint != endpoint:
        return False

    if request.url_rule is None:  # pragma: no cover
        return False

    try:
        rule_url = request.url_rule.build(url_kwargs, append_unknown=not ignore_query)
    except TypeError:  # pragma: no cover
        return False

    if rule_url is None:  # pragma: no cover
        return False

    _, url = rule_url

    return url == request.path


def is_active_blueprint(blueprint: str) -> bool:
    """Check if the current request is for the given blueprint"""
    return request.blueprint == blueprint



from collections.abc import Iterator
from pathlib import PosixPath as Path

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.util import container

from bootlace.endpoint import Endpoint


@attrs.define
class Resources:
    """Bootstrap resource extension for bootlace"""

    endpoint: Endpoint

    resources: list[str] = attrs.field(factory=list)

    def iter_resources(self, extension: str | None = None) -> Iterator[str]:
        if extension and not extension.startswith("."):
            extension = f".{extension}"

        for resource in self.resources:
            if extension and Path(resource).suffix == extension:
                yield resource

    def css(self) -> dom_tag:
        collection = container()

        for filename in self.iter_resources(".css"):
            url = self.endpoint(filename=filename)
            collection.add(tags.link(rel="stylesheet", href=url))

        return collection

    def js(self) -> dom_tag:
        collection = container()

        for filename in self.iter_resources(".js"):
            url = self.endpoint(filename=filename)
            collection.add(tags.script(src=url))

        return collection



from collections.abc import Iterator
from collections.abc import Mapping
from typing import Any
from typing import Protocol

import attrs
from flask import Blueprint
from flask import request
from flask import url_for
from werkzeug.exceptions import BadRequest
from werkzeug.routing import BuildError

from bootlace.util import is_active_endpoint


class NoEndpointError(BadRequest):
    """Raised when there is no endpoint set."""


@attrs.define(frozen=True, init=False)
class KeywordArguments(Mapping[str, Any]):
    """
    A mapping of keyword arguments for a URL endpoint,
    which is frozen and hashable for use as a key in a dictionary.
    """

    _arguments: frozenset[tuple[str, Any]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        arguments = frozenset(dict(*args, **kwargs).items())
        object.__setattr__(self, "_arguments", arguments)

    def as_dict(self) -> dict[str, Any]:
        return dict(self._arguments)

    def __getitem__(self, __key: str) -> Any:
        return self.as_dict()[__key]

    def __iter__(self) -> Iterator[str]:
        return iter((key for key, _ in self._arguments))

    def __len__(self) -> int:
        return len(self._arguments)

    def __repr__(self) -> str:
        return f"KeywordArguments({self.as_dict()!r})"


class EndpointProtocol(Protocol):
    @property
    def active(self) -> bool: ...

    def build(self, **kwds: Any) -> str: ...

    def __call__(self, **kwds: Any) -> str: ...

    @property
    def name(self) -> str: ...

    @property
    def full_name(self) -> str: ...

    @property
    def url(self) -> str: ...

    @property
    def blueprint(self) -> str | None: ...

    @property
    def url_kwargs(self) -> KeywordArguments: ...

    @property
    def ignore_query(self) -> bool: ...


@attrs.define(frozen=True, repr=False, kw_only=True)
class Endpoint:
    """An endpoint for a breadcrumb, as captured at registration"""

    #: The flask context, if any
    context: None | Blueprint = None

    #: The endpoint name
    name: str

    #: The keyword arguments for the endpoint used with url_for
    url_kwargs: KeywordArguments = attrs.field(factory=lambda: KeywordArguments(), converter=KeywordArguments)

    #: Whether to ignore the query string when checking for active status
    ignore_query: bool = True

    @classmethod
    def from_name(cls, name: str, **kwargs: Any) -> "Endpoint":
        """Create an endpoint from a name and keyword arguments"""
        ignore_query = kwargs.pop("ignore_query", True)

        return cls(name=name, url_kwargs=KeywordArguments(**kwargs), ignore_query=ignore_query)

    @property
    def full_name(self) -> str:
        """The full name of the endpoint"""
        if self.context is not None and "." not in self.name:
            return f"{self.context.name}.{self.name}"

        return self.name

    @property
    def blueprint(self) -> str | None:
        """The blueprint for the endpoint"""
        if "." in self.name:
            return ".".join(self.name.split(".")[:-1])

        if self.context is not None:
            return self.context.name

        return None

    @property
    def url(self) -> str:
        """The URL for the endpoint"""
        return self.build()

    def build(self, **kwds: Any) -> str:
        """Build the URL for the endpoint with additional keyword arguments"""
        if isinstance(self.context, Blueprint):
            name = f"{self.context.name}.{self.name}"
            return url_for(name, **self.url_kwargs, **kwds)

        return url_for(self.name, **self.url_kwargs, **kwds)

    @property
    def active(self) -> bool:
        """Whether the endpoint is active"""
        return is_active_endpoint(self.full_name, self.url_kwargs, self.ignore_query)

    def __call__(self, **kwds: Any) -> str:
        return self.build(**kwds)

    def __repr__(self) -> str:
        parts = []
        if self.context is not None:
            parts.append(f"{self.context.name:s}.{self.name:s}")
        else:
            parts.append(f"{self.name:s}")

        if self.url_kwargs:
            parts.append(f", {self.url_kwargs!r}")

        if not self.ignore_query:
            parts.append(", ignore_query=False")

        statement = ", ".join(parts)
        return f"Endpoint({statement})"


@attrs.define(frozen=True)
class CurrentEndpoint:
    ignore_query: bool = False

    @property
    def active(self) -> bool:
        return True

    @property
    def name(self) -> str:
        if not request.endpoint:
            raise NoEndpointError()

        return request.endpoint.split(".")[-1]

    @property
    def full_name(self) -> str:
        if not request.endpoint:
            raise NoEndpointError()
        return request.endpoint

    @property
    def blueprint(self) -> str | None:
        return request.blueprint

    @property
    def url(self) -> str:
        return self.build()

    @property
    def url_kwargs(self) -> KeywordArguments:
        args = dict(request.view_args or {})
        args.update(request.args.items())
        return KeywordArguments(args)

    def build(self, **kwds: Any) -> str:
        args = request.view_args or {}
        args.update(kwds)

        if not request.endpoint:
            raise BuildError(request.endpoint, args, request.method)

        args["_method"] = request.method
        return url_for(request.endpoint, **args)

    def __call__(self, **kwds: Any) -> str:
        return self.build(**kwds)


def convert_endpoint(endpoint: str | Endpoint) -> Endpoint:
    """Convert a string endpoint to an Endpoint object"""
    if isinstance(endpoint, str):
        return Endpoint.from_name(endpoint)

    return endpoint



from .base import ColumnBase
from .base import Heading
from .base import Table
from .columns import ActionColumn
from .columns import CheckColumn
from .columns import Column
from .columns import Datetime

__all__ = [
    "Table",
    "ColumnBase",
    "Heading",
    "Column",
    "CheckColumn",
    "Datetime",
    "ActionColumn",
]



import warnings

import attrs
from domilite import tags
from marshmallow import fields

from bootlace.util import BootlaceWarning
from bootlace.util import MaybeTaggable
from bootlace.util import Tag
from bootlace.util import as_tag
from bootlace.util import ids as element_id

from .core import NavAlignment
from .core import NavStyle
from .core import SubGroup


@attrs.define
class Nav(SubGroup):
    """A navigation bar"""

    #: The ID of the nav
    id: str = attrs.field(factory=element_id.factory("nav"))

    #: The style of the nav
    style: NavStyle = attrs.field(default=NavStyle.PLAIN, metadata={"form": fields.Enum(NavStyle)})

    #: The alignment of the elments in the nav
    alignment: NavAlignment = attrs.field(default=NavAlignment.DEFAULT, metadata={"form": fields.Enum(NavAlignment)})

    ul: Tag = Tag(tags.ul, classes={"nav"})
    li: Tag = Tag(tags.li, classes={"nav-item"})

    def __tag__(self) -> tags.html_tag:
        active_endpoint = next((item for item in self.items if item.active), None)
        ul = self.ul(id=self.id)

        if (style := self.style.value) != "":
            ul.classes.add(style)

        if (alignment := self.alignment.value) != "":
            ul.classes.add(alignment)

        if (link := getattr(active_endpoint, "link", None)) is not None:
            if (endpoint := getattr(link, "endpoint", None)) is not None:
                ul["data-endpoint"] = endpoint.full_name

        for item in self.items:
            ul.add(self.li(as_tag(item), __pretty=False))

        return ul


@attrs.define
class Dropdown(SubGroup):
    """A dropdown menu in the nav bar"""

    #: The title of the dropdown
    title: MaybeTaggable = attrs.field(kw_only=True)

    #: The ID of the dropdown
    id: str = attrs.field(factory=element_id.factory("bs-dropdown"))

    dropdown: Tag = Tag(tags.div, classes={"dropdown"})
    ul: Tag = Tag(tags.ul, classes={"dropdown-menu"})
    li: Tag = Tag(tags.li)
    toggle: Tag = Tag(tags.a, classes={"nav-link", "dropdown-toggle"}, attributes={"role": "button"})

    def __tag__(self) -> tags.html_tag:
        div = self.dropdown()
        a = self.toggle(
            as_tag(self.title),
            href="#",
            id=self.id,
        )
        a.aria["expanded"] = "false"
        a.data["bs-toggle"] = "dropdown"
        div.add(a)

        menu = self.ul()
        menu.aria["labelledby"] = self.id
        for item in self.items:
            tag = as_tag(item)
            if isinstance(tag, tags.html_tag):
                tag.classes.discard("nav-link")
                if not any(cls.startswith("dropdown-") for cls in tag.classes):
                    tag.classes.add("dropdown-item")
            else:
                warnings.warn(
                    BootlaceWarning(f"Item {item!r} is not an html tag, may not display properly"),
                    stacklevel=2,
                )
            menu.add(self.li(tag, __pretty=False))
        div.add(menu)
        return div



import inspect
from abc import ABC
from abc import abstractmethod
from collections.abc import Iterable
from collections.abc import Mapping
from typing import Any
from typing import ClassVar

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.util import text

from bootlace.icon import Icon
from bootlace.util import Tag
from bootlace.util import as_tag
from bootlace.util import maybe


@attrs.define
class Heading:
    """A heading for a table column."""

    #: The text of the heading
    text: str

    #: The icon for the heading, in place of the text
    icon: Icon | None = attrs.field(default=None, converter=maybe(Icon))  # type: ignore

    def __tag__(self) -> tags.html_tag:
        if self.icon:
            return tags.a(
                as_tag(self.icon),
                href="#",
                data_bs_toggle="tooltip",
                data_bs_title=self.text,
                cls="link-dark",
            )
        return tags.span(self.text)


@attrs.define
class ColumnBase(ABC):
    """Base class for table columns.

    Subclasses must implement the :meth:`cell` method."""

    #: The heading for the column
    heading: Heading = attrs.field(converter=maybe(Heading))  # type: ignore

    name: str | None = None

    th = Tag(tags.th)
    td = Tag(tags.td)

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = self.name or name

    @property
    def attribute(self) -> str:
        """The attribute name for the column."""
        if self.name is None:
            raise ValueError("column must be named in Table or name= parameter must be provided")
        return self.name

    def attribute_value(self, value: Any) -> Any:
        """Return the value of the attribute for the given object."""
        return getattr(value, self.attribute)

    def contents(self, value: Any, format: str | None = None) -> Any:
        """Return the contents of the cell for the column, using an HTML comment if the attribute value is None."""
        contents = self.attribute_value(value)
        if contents is None:
            return tags.comment(f"No value for {self.name}")
        if format:
            return text(format.format(contents))
        return text(str(contents))

    @abstractmethod
    def cell(self, value: Any) -> dom_tag:
        """Return the cell for the column as an HTML tag."""
        raise NotImplementedError("Subclasses must implement this method")

    def __th__(self) -> dom_tag:
        return self.th(as_tag(self.heading), scope="col", __pretty=False)

    def __td__(self, value: Any) -> dom_tag:
        return self.td(self.cell(value), __pretty=False)


def is_instance_or_subclass(val: Any, class_: type) -> bool:
    """Return True if ``val`` is either a subclass or instance of ``class_``."""
    try:
        return issubclass(val, class_)
    except TypeError:
        return isinstance(val, class_)


def _get_columns(attrs: Mapping[str, Any]) -> dict[str, ColumnBase]:
    return {
        column_name: column_value
        for column_name, column_value in attrs.items()
        if is_instance_or_subclass(column_value, ColumnBase)
    }


class TableMetaclass(type):
    columns: dict[str, ColumnBase]

    def __new__(mcls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]) -> type:
        cls = super().__new__(mcls, name, bases, namespace)
        cls.columns = mcls.get_declared_columns(cls)
        cls.columns.update(_get_columns(namespace))
        return cls

    @classmethod
    def get_declared_columns(mcls, cls: type) -> dict[str, ColumnBase]:
        mro = inspect.getmro(cls)
        # Loop over mro in reverse to maintain correct order of fields
        columns: dict[str, ColumnBase] = {}

        column_gen = (
            _get_columns(
                getattr(base, "_declared_columns", base.__dict__),
            )
            for base in mro[:0:-1]
        )

        for column_set in column_gen:
            columns.update(column_set)

        return columns


class Table(metaclass=TableMetaclass):
    """Base class for class-defined tables.

    Subclasses should define columns as class attributes, e.g.:

    class MyTable(Table):
        name = Column(Heading("Name"))
        age = Column(Heading("Age"))

    Use :meth:`render` to render a table from a list of items as
    :class:`dominate.tags.table`.
    """

    columns: ClassVar[dict[str, ColumnBase]]

    table = Tag(tags.table)
    thead = Tag(tags.thead)
    tbody = Tag(tags.tbody)
    tr = Tag(tags.tr)

    def __init__(self, decorated_classes: Iterable[str] | None = None) -> None:
        if decorated_classes is None:
            self.decorated_classes = set()
        else:
            self.decorated_classes = set(decorated_classes)

    def __table__(self, items: list[Any]) -> tags.html_tag:
        table = self.table(cls="table")
        table.classes.add(*self.decorated_classes)
        table.add(self.__thead__())
        table.add(self.__tbody__(items))
        return table

    def __thead__(self) -> tags.html_tag:
        thead = self.thead()
        for column in self.columns.values():
            thead.add(column.__th__())
        return thead

    def __tr__(self, item: Any) -> tags.html_tag:
        id = getattr(item, "id", None)
        name = item.__class__.__name__.lower()
        tr = self.tr(id=f"{name}-{id}" if id else None)
        for column in self.columns.values():
            tr.add(column.__td__(item))
        return tr

    def __tbody__(self, items: list[Any]) -> tags.html_tag:
        tbody = self.tbody()
        for item in items:
            tbody.add(self.__tr__(item))
        return tbody

    def __call__(self, items: list[Any]) -> tags.html_tag:
        return self.__table__(items)



import warnings
from typing import Any
from typing import ClassVar

import attrs
from domilite import tags
from markupsafe import Markup
from markupsafe import escape
from wtforms.fields import Field

from bootlace.util import Tag
from bootlace.util import render

__all__ = [
    "Widget",
    "Input",
    "TextInput",
    "PasswordInput",
    "HiddenInput",
    "CheckboxInput",
    "RadioInput",
    "FileInput",
    "SubmitInput",
    "TextArea",
]


@attrs.define
class Widget:
    classes: ClassVar[set[str]]
    tag: ClassVar[Tag]
    validation_attrs: ClassVar[set[str]] = set()

    def get_field_value(self, field: Field) -> Any:
        if (get_value := getattr(field, "_value", None)) is not None:
            return get_value()
        return field.data

    def prepare_attributes(self, field: Field, kwargs: dict[str, Any]) -> dict[str, Any]:
        return kwargs

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        kwargs.setdefault("id", field.id)
        if "value" not in kwargs and ((value := self.get_field_value(field)) is not None):
            kwargs["value"] = value

        flags = getattr(field, "flags", object())
        for k in dir(flags):
            if k in self.validation_attrs and k not in kwargs:
                kwargs[k] = getattr(flags, k)

        kwargs = self.prepare_attributes(field, kwargs)
        return self.tag(**kwargs)

    def __call__(self, field: Field, **kwargs: Any) -> Markup:
        return render(self.__form_tag__(field, **kwargs))


@attrs.define
class InputBase(Widget):
    """A base input widget"""

    input_type: ClassVar[str | None] = None
    tag = Tag(tags.input, classes={"form-control"})

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if self.input_type is not None:
            kwargs.setdefault("type", self.input_type)
        else:
            warnings.warn("Input type not specified", stacklevel=2)
        return super().__form_tag__(field, **kwargs)


@attrs.define
class Input(Widget):
    """A base input widget"""

    input_type: str | None = None
    tag = Tag(tags.input, classes={"form-control"})

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if self.input_type is not None:
            kwargs.setdefault("type", self.input_type)
        else:
            warnings.warn("Input type not specified", stacklevel=2)
        return super().__form_tag__(field, **kwargs)


@attrs.define
class TextInput(InputBase):
    """Render a single-line text input"""

    input_type = "text"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }


@attrs.define
class PasswordInput(InputBase):
    hide_value: bool = True

    input_type = "password"
    validation_attrs = {
        "required",
        "disabled",
        "readonly",
        "maxlength",
        "minlength",
        "pattern",
    }

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if self.hide_value:
            kwargs["value"] = ""
        return super().__form_tag__(field, **kwargs)


@attrs.define
class HiddenInput(InputBase):
    input_type = "hidden"
    validation_attrs = {"disabled"}


@attrs.define
class CheckboxInput(InputBase):
    input_type = "checkbox"
    validation_attrs = {"required", "disabled"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if getattr(field, "checked", field.data):
            kwargs["checked"] = True
        return super().__form_tag__(field, **kwargs)


@attrs.define
class RadioInput(InputBase):
    input_type = "radio"
    validation_attrs = {"required", "disabled"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if getattr(field, "checked", field.data):
            kwargs["checked"] = True
        return super().__form_tag__(field, **kwargs)


@attrs.define
class FileInput(InputBase):
    multiple: bool = False

    input_type = "file"
    validation_attrs = {"required", "disabled", "accept"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        kwargs["value"] = False
        if self.multiple:
            kwargs["multiple"] = True

        return super().__form_tag__(field, **kwargs)


@attrs.define
class SubmitInput(InputBase):
    input_type = "submit"
    validation_attrs = {"required", "disabled"}

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        kwargs.setdefault("value", field.label.text)
        return super().__form_tag__(field, **kwargs)


@attrs.define
class TextArea(Widget):
    validation_attrs = {"required", "disabled", "readonly", "maxlength", "minlength"}

    tag = Tag(tags.textarea, classes={"form-control"})

    def get_field_value(self, field: Field) -> Any:
        value = super().get_field_value(field)
        if value is not None:
            return escape(value)
        return value



import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.util import container
from marshmallow import fields

from bootlace.size import SizeClass
from bootlace.style import ColorClass
from bootlace.util import Tag
from bootlace.util import as_tag
from bootlace.util import ids as element_id

from .core import Link
from .core import NavElement
from .core import SubGroup
from .nav import Nav
from .schema import NavSchema


@attrs.define
class NavBar(NavElement):
    """A navigation bar, typically at the top of the page

    This is usually the primary navigation for a site.
    """

    #: The ID of the navbar
    id: str = attrs.field(factory=element_id.factory("navbar"))

    #: The elements in the navbar
    items: list[NavElement] = attrs.field(factory=list, metadata={"form": fields.List(fields.Nested(NavSchema))})

    #: The size of the navbar, if any, used to select when it
    #: should expand or collapse
    expand: SizeClass | None = attrs.field(
        default=SizeClass.LARGE,
        metadata={"form": fields.Enum(SizeClass, allow_none=True)},
    )

    #: The color of the navbar, if using a bootstrap color class
    color: ColorClass | None = attrs.field(
        default=ColorClass.TERTIARY,
        metadata={"form": fields.Enum(ColorClass, allow_none=True)},
    )

    #: Whether the navbar should be fluid (e.g. full width)
    fluid: bool = True

    nav: Tag = Tag(tags.nav, classes={"navbar"})
    container: Tag = Tag(tags.div, classes={"container"})

    def __tag__(self) -> tags.html_tag:
        nav = self.nav()
        if self.expand:
            nav.classes.add(self.expand.add_to_class("navbar-expand"))
        if self.color:
            nav.classes.add(self.color.add_to_class("bg"))

        container = self.container()
        if self.fluid:
            container.classes.add("container-fluid")
            container.classes.discard("container")
        else:
            container.classes.add("container")

        nav.add(container)

        for item in self.items:
            container.add(as_tag(item))

        return nav


@attrs.define
class Brand(Link):
    """The brand for the navbar, typically the site's logo or name

    You can pass :class:`~bootlace.links.Link` or :class:`~bootlace.links.View`
    as the source link, and
    """

    #: The ID of the brand
    id: str = attrs.field(factory=element_id.factory("navbar-brand"))

    def __tag__(self) -> dom_tag:
        a = as_tag(self.link)
        a["class"] = "navbar-brand"
        a["id"] = self.id
        return self.element_state(a)


@attrs.define
class NavBarCollapse(SubGroup):
    """A collection of nav elements that can be collapsed"""

    id: str = attrs.field(factory=element_id.factory("navbar-collapse"))

    button: Tag = Tag(tags.button, classes={"navbar-toggler"}, attributes={"type": "button"})
    icon: Tag = Tag(tags.span, classes={"navbar-toggler-icon"})
    container: Tag = Tag(tags.div, classes={"collapse", "navbar-collapse"})

    def __tag__(self) -> dom_tag:
        button = self.button()
        button.data["bs-toggle"] = "collapse"
        button.data["bs-target"] = f"#{self.id}"
        button.aria["controls"] = f"{self.id}"
        button.aria["expanded"] = "false"
        button.aria["label"] = "Toggle navigation"

        button.add(self.icon())
        div = self.container(id=self.id)
        for item in self.items:
            div.add(as_tag(item))
        return container(button, div)


@attrs.define
class NavBarNav(Nav):
    """Primary grouping of nav elements in the navbar"""

    id: str = attrs.field(factory=element_id.factory("navbar-nav"))

    ul: Tag = Tag(tags.ul, classes={"navbar-nav"})
    li: Tag = Tag(tags.li, classes={"nav-item"})

    def __tag__(self) -> tags.html_tag:
        ul = self.ul(id=self.id)
        for item in self.items:
            ul.add(self.li(as_tag(item), __pretty=False))
        return ul


@attrs.define
class NavBarSearch(NavElement):
    """A search bar for the navbar"""

    id: str = attrs.field(factory=element_id.factory("navbar-search"))

    placeholder: str = "Search"
    action: str = "#"
    method: str = "GET"
    button: str | None = None

    form: Tag = Tag(tags.form, classes={"d-flex"}, attributes={"role": "search"})
    input: Tag = Tag(tags.input, classes={"form-control", "me-2"}, attributes={"type": "search"})
    button_tag: Tag = Tag(tags.button, classes={"btn", "btn-success"}, attributes={"type": "submit"})

    def __tag__(self) -> dom_tag:
        form = self.form(id=self.id)

        input = self.input(
            placeholder=self.placeholder,
        )
        input.aria["label"] = self.placeholder
        form.add(input)
        form.add(self.button_tag(self.button or self.placeholder))
        return self.element_state(form)



from typing import Any

import attrs
from domilite import tags
from wtforms.fields import Field

from bootlace.util import Tag

from .core import Widget


@attrs.define
class Switch(Widget):
    div: Tag = Tag(tags.div, classes={"form-check", "form-switch"})
    input: Tag = Tag(
        tags.input,
        classes={"form-check-input"},
        attributes={"type": "checkbox", "role": "switch"},
    )
    label: Tag = Tag(tags.label, classes={"form-check-label"})

    def __form_tag__(self, field: Field, **kwargs: Any) -> tags.html_tag:
        if getattr(field, "checked", field.data):
            kwargs["checked"] = True

        div = self.div()
        div.add(self.input(**kwargs))
        label = div.add(self.label(field.label.text))
        label["for"] = field.id
        return div



import textwrap
from collections.abc import Callable
from xml.etree import ElementTree

import attrs
import html5lib

Element = ElementTree.Element

Filter = Callable[[Element], Element | None]

ATTRIBUTES = {
    "role",
}


@attrs.define(str=False)
class Difference:
    expected: str | None
    actual: str | None
    message: str

    def __str__(self) -> str:
        return f"{self.message}: \n Expected: {self.expected}\n Found: {self.actual}"


@attrs.define(str=False)
class TagDifference:
    expected: Element
    actual: Element | None
    differences: list[Difference] = attrs.field(factory=list)

    def __str__(self) -> str:
        differences = textwrap.indent("\n".join(str(difference) for difference in self.differences), prefix="  ")
        return f"Tag {self.expected.tag} differences:\n{differences}"

    def __bool__(self) -> bool:
        return bool(self.differences)

    def append(self, difference: Difference) -> None:
        self.differences.append(difference)


@attrs.define(str=False)
class Result:
    differences: list[TagDifference] = attrs.field(factory=list)

    def __bool__(self) -> bool:
        return bool(self.differences)

    def __str__(self) -> str:
        if not self.differences:
            return "No differences found"
        else:
            n = sum(len(differences.differences) for differences in self.differences)
            message = f"{n} differences found"
            details = "\n".join(textwrap.indent(str(difference), prefix="  ") for difference in self.differences)
            return f"{message}:\n{details}"


@attrs.define
class HTMLDiff:
    filters: list[Filter] = attrs.field(factory=list)
    attributes: set[str] = attrs.field(factory=set)

    def compare(self, expected_html: str, actual_html: str) -> Result:
        expected = html5lib.parse(expected_html, namespaceHTMLElements=False)
        actual = html5lib.parse(actual_html, namespaceHTMLElements=False)

        differences: list[TagDifference] = []
        actual = actual.iter()
        for element in expected.iter():
            element = self._filter(element)
            if element is None:
                continue

            tagdiff = self._compare(element, next(actual, None))
            if tagdiff:
                differences.append(tagdiff)
            if tagdiff.actual is None:
                # Ran out of tags to compare
                break

        return Result(differences)

    def _filter(self, element: Element) -> Element | None:
        for filter in self.filters:
            element = filter(element)  # type: ignore
            if element is None:
                return None
        return element

    def _compare(self, expected: Element, actual: Element | None) -> TagDifference:
        if actual is None:
            return TagDifference(expected, actual, [Difference(expected.tag, None, "Tag not found")])

        differences = TagDifference(expected, actual)
        if expected.tag != actual.tag:
            differences.append(Difference(expected.tag, actual.tag, "Tag mismatch"))

        classes = set(expected.get("class", "").split())
        actual_classes = set(actual.get("class", "").split())
        if classes != actual_classes:
            differences.append(Difference(",".join(classes), ",".join(actual_classes), "Class mismatch"))

        if "href" in expected.attrib:
            expected_href = expected.attrib["href"]
            actual_href = actual.attrib.get("href", "")

            if expected_href == "#":
                expected_href = ""
            if actual_href == "#":
                actual_href = ""
            if expected_href != actual_href:
                differences.append(Difference(expected_href, actual_href, "Href mismatch"))

        for attr, value in expected.attrib.items():
            if (attr in self.attributes) or (attr.startswith("aria-")):
                if attr not in actual.attrib:
                    differences.append(Difference(attr, None, "Attribute not found"))
                    continue
                if actual.get(attr) != value:
                    differences.append(
                        Difference(
                            value,
                            actual.get(attr),
                            f"Attribute {attr} mismatch",
                        )
                    )
                    continue

        for attr in actual.attrib:
            if (attr in self.attributes) or (attr.startswith("aria-")):
                if attr not in expected.attrib:
                    differences.append(
                        Difference(
                            None,
                            actual.attrib.get(attr),
                            f"Attribute {attr} not expected",
                        )
                    )
                    continue

        return differences


def assert_same_html(expected_html: str, actual_html: str) -> None:
    __tracebackhide__ = True
    diff = HTMLDiff(attributes=ATTRIBUTES)
    result = diff.compare(expected_html, actual_html)
    assert not result, str(result)



import enum


class SizeClass(enum.Enum):
    """Bootstrap size classes"""

    #: Extra-small is the default size class used when none is specified. < 576px
    EXTRA_SMALL = None

    #: Small size class <768px
    SMALL = "sm"

    #: Medium size class <992px
    MEDIUM = "md"

    #: Large size class <1200px
    LARGE = "lg"

    #: Extra-large size class <1400px
    EXTRA_LARGE = "xl"

    #: Extra-extra-large size class >=1400px
    EXTRA_EXTRA_LARGE = "xxl"

    def add_to_class(self, cls: str) -> str:
        """Add the size class to the given class."""
        if self.value:
            return f"{cls}-{self.value}"
        else:
            return cls



from collections.abc import Callable
from collections.abc import Iterator
from typing import Protocol
from typing import TypeVar

import attrs
from domilite import tags
from domilite.dom_tag import dom_tag
from domilite.util import text
from flask import Blueprint
from flask import Flask
from flask import current_app
from flask import request
from werkzeug.local import LocalProxy

from bootlace.endpoint import Endpoint

from .util import Tag
from .util import as_tag

__all__ = [
    "breadcrumbs",
    "Breadcrumb",
    "Breadcrumbs",
    "BreadcrumbEntry",
    "BreadcrumbExtension",
]


class Named(Protocol):
    __name__: str


V = TypeVar("V", bound=Named)

EXTENSION_KEY: str = "bootlace.breadcrumbs"
DIVIDER_SETTING: str = "BOOTLACE_BREADCRUMBS_DIVIDER"


@attrs.define
class Breadcrumb:
    """A single breadcrumb, representing a single, possibly active, endpoint"""

    #: The title of the breadcrumb, displayed in text
    title: str

    #: The endpoint for the breadcrumb
    link: Endpoint

    a: Tag = Tag(tags.a)

    @property
    def active(self) -> bool:
        """Whether the breadcrumb is the active view"""
        return self.link.active

    @property
    def url(self) -> str:
        """The URL for the breadcrumb"""
        return self.link.url

    def __tag__(self) -> dom_tag:
        if self.active:
            return text(self.title)

        return self.a(self.title, href=self.url)


@attrs.define
class Breadcrumbs:
    """The trail of breadcrumbs

    Supports the :func:`as_tag` protocol to render the breadcrumbs as HTML"""

    #: The list of breadcrumbs, in order from broadest to most specific
    crumbs: list[Breadcrumb] = attrs.field(factory=list)

    #: The divider to use between breadcrumbs
    divider: str = attrs.field(default=">")

    nav: Tag = Tag(tags.nav)
    ol: Tag = Tag(tags.ol, classes={"breadcrumb"})
    li: Tag = Tag(tags.li, classes={"breadcrumb-item"})

    def __iter__(self) -> Iterator[Breadcrumb]:
        return iter(self.crumbs)

    def __len__(self) -> int:
        return len(self.crumbs)

    def __getitem__(self, index: int) -> Breadcrumb:
        return self.crumbs[index]

    def push(self, crumb: Breadcrumb) -> None:
        """Add a new crumb to the beginning of the list"""
        self.crumbs.insert(0, crumb)

    def __tag__(self) -> dom_tag:
        if not self.crumbs:
            return text("")

        nav = self.nav()
        nav.aria["label"] = "breadcrumb"

        if self.divider != "/":
            nav["style"] = f"--breadcrumb-divider: '{self.divider:s}';"  # noqa: B907

        ol = self.ol()
        for crumb in self:
            item = self.li(as_tag(crumb))
            if crumb.active:
                item.aria["current"] = "page"
                item.classes.add("active")
            ol.add(item)
        nav.add(ol)
        return nav


@attrs.define
class BreadcrumbEntry:
    """A single entry in the breadcrumbs datastructure"""

    title: str
    parent: Endpoint | None


@attrs.define(init=False)
class BreadcrumbExtension:
    """An extension for breadcrumbs"""

    tree: dict[Endpoint, BreadcrumbEntry] = attrs.field(factory=dict)

    def __init__(self, app: Flask | None = None) -> None:
        self.tree = {}
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Set up the extension on the app"""
        app.config.setdefault(DIVIDER_SETTING, ">")
        app.extensions[EXTENSION_KEY] = self

    def register(
        self,
        context: Flask | Blueprint | None,
        parent: str | Endpoint | None,
        title: str,
    ) -> Callable[[V], V]:
        """Register a breadcrumb for a view

        :param context: The context for the view, if any.
            For a blueprint, this is the blueprint object. For a view on the root app,
            this ie either the root app or None.
        :param parent: The parent for the breadcrumb. This can be a string, an endpoint, or None.
            If a string, it is the name of the parent endpoint. If an endpoint, it is the parent endpoint.
            If None, there is no parent, and this is a root item.
        :param title: The title of the breadcrumb

        :returns: A decorator that can be used to register the view, which will return the view unchanged.
        """
        if isinstance(context, Flask):
            context = None

        parent_link: Endpoint | None = None
        if isinstance(parent, str):
            if parent.startswith("."):
                if context is None:
                    raise ValueError("Cannot use relative endpoint without a context")
                parent_link = Endpoint(name=parent.lstrip("."), context=context)
            else:
                parent_link = Endpoint(name=parent, context=None)
        else:
            parent_link = parent

        def decorator(view: V) -> V:
            nonlocal parent_link
            link = Endpoint(name=view.__name__, context=context)

            if link == parent_link:
                raise ValueError("A breadcrumb cannot be its own parent")

            self.tree[link] = BreadcrumbEntry(title=title, parent=parent_link)
            return view

        return decorator

    @property
    def divider(self) -> str:
        """The divider to use between breadcrumbs"""
        return current_app.config[DIVIDER_SETTING]

    def _current_context(self) -> Blueprint | None:
        if request.blueprint:
            return current_app.blueprints[request.blueprint]  # type: ignore
        return None

    def _current_endpoint(self) -> Endpoint | None:
        context = self._current_context()
        if not request.endpoint:  # pragma: no cover
            return None

        name = request.endpoint.split(".")[-1]

        return Endpoint(name=name, context=context)

    def get(self) -> Breadcrumbs:
        """Get the :class:`Breadcrumbs` for the current request"""
        endpoint = self._current_endpoint()
        crumbs = Breadcrumbs(divider=self.divider)
        if not endpoint:  # pragma: no cover
            return crumbs

        current = self.tree.get(endpoint)
        while current:
            crumbs.push(Breadcrumb(title=current.title, link=endpoint))
            if not current.parent:
                break
            endpoint = current.parent
            current = self.tree.get(endpoint)

        return crumbs


breadcrumbs: BreadcrumbExtension = LocalProxy(lambda: current_app.extensions[EXTENSION_KEY])  # type: ignore



import enum
import mimetypes
from typing import Any
from typing import Generic
from typing import TypeVar

from werkzeug.http import parse_options_header
from wtforms import fields
from wtforms.validators import Regexp
from wtforms.validators import ValidationError

__all__ = ["EnumField", "MarkdownField", "KnownMIMEType", "SLUG_VALIDATOR"]


SLUG_VALIDATOR = Regexp(r"^[A-Za-z0-9\-]+$", message="Slug must be a URL path component")


E = TypeVar("E", bound=enum.Enum)


def _import_fields() -> None:
    for name in dir(fields):
        if not name.startswith("_") and name[0].isupper():
            globals()[name] = getattr(fields, name)
            __all__.append(name)


def _enum_labelfunc(value: E) -> str:
    return value.name.capitalize().replace("_", "-")


class EnumField(fields.SelectField, Generic[E]):
    """Field for selecting an enum value. The choices are generated from the enum values. The label is the enum value.

    :param label: The label for the field
    :param validators: Validators for the field
    :param enum: The enum type
    """

    def __init__(
        self,
        label: str | None = None,
        validators: Any = None,
        *,
        enum: type[E],
        **kwargs: Any,
    ) -> None:
        labelfunc = kwargs.pop("labelfunc", _enum_labelfunc)
        kwargs.setdefault("choices", [(value.name, labelfunc(value)) for value in enum])
        super().__init__(label=label, validators=validators, coerce=self._coerce, **kwargs)
        self.enum = enum

    def _coerce(self, value: str | E) -> E:
        if isinstance(value, self.enum):
            return value
        return self.enum[value]  # type: ignore

    def _value(self) -> str:
        if self.data:
            return self.data.name
        else:
            return ""


def unwrap_paragraphs(txt: str) -> str:
    return "\n\n".join([paragraph.replace("\n", " ") for paragraph in txt.split("\n\n")])


class MarkdownField(fields.TextAreaField):
    """TextArea field designed for Markdown content.

    The field will unwrap paragraphs and remove newlines from paragraphs."""

    def _value(self) -> str:
        if self.data:
            return unwrap_paragraphs(self.data)
        else:
            return ""


class KnownMIMEType:
    """Validator that ensures that a content-type field is a known MIME type"""

    def __init__(self, message: str | None = None) -> None:
        if not message:
            message = "Must be a well known MIME type."
        self.message = message
        self.db = mimetypes.MimeTypes()

    def __call__(self, form: Any, field: fields.Field) -> None:
        """Validates the field

        :param form: The form contianing the field
        :param field: The field

        :raises ValidationError: If the field is not a known MIME type
        """
        mimetype, _options = parse_options_header(field.data)

        well_known_types, official_types = self.db.types_map_inv

        if (mimetype not in well_known_types) and (mimetype not in official_types):
            raise ValidationError(self.message)


_import_fields()



import attrs
from domilite import tags

__all__ = ["Image"]


@attrs.define(kw_only=True, frozen=True)
class Image:
    """An image tag"""

    #: Alt text for the image
    alt: str

    #: The URL for the image
    src: str

    #: The width of the image, in px
    width: int

    #: The height of the image, in px
    height: int

    def __tag__(self) -> tags.html_tag:
        return tags.img(src=self.src, alt=self.alt, width=str(self.width), height=str(self.height))



from typing import Any

from markupsafe import Markup
from wtforms import widgets as wt_widgets
from wtforms.fields import Field
from wtforms.meta import DefaultMeta

from . import widgets as bl_widgets


class BootlaceMeta(DefaultMeta):
    WIDGET_MAP = {getattr(wt_widgets, name): getattr(bl_widgets, name) for name in bl_widgets.__all__}

    def render_field(self, field: Field, render_kw: Any) -> Markup:
        cls = self.WIDGET_MAP[field.widget.__class__]
        widget = cls()
        return widget(field, render_kw)



from .bar import Brand
from .bar import NavBar
from .bar import NavBarCollapse
from .bar import NavBarNav
from .bar import NavBarSearch
from .core import Link
from .core import Separator
from .core import Text
from .nav import Dropdown
from .nav import Nav

__all__ = [
    "Link",
    "Separator",
    "Text",
    "NavBar",
    "Brand",
    "NavBarCollapse",
    "NavBarNav",
    "NavBarSearch",
    "Nav",
    "Dropdown",
]
